@@ -213,9 +213,16 @@ typedef struct _xf86CrtcFuncs {
     Bool
     (*set_mode_major)(xf86CrtcPtr crtc, DisplayModePtr mode,
 		      Rotation rotation, int x, int y);
+
+    /**
+     * Callback for panning. Doesn't change the mode.
+     */
+    void
+    (*set_origin)(xf86CrtcPtr crtc, int x, int y);
+
 } xf86CrtcFuncsRec, *xf86CrtcFuncsPtr;
 
-#define XF86_CRTC_VERSION 1
+#define XF86_CRTC_VERSION 2
 
 struct _xf86Crtc {
     /**
@@ -305,12 +312,31 @@ struct _xf86Crtc {
      * Current transformation matrix
      */
     PictTransform   crtc_to_framebuffer;
-    PictTransform   framebuffer_to_crtc;
+    struct pict_f_transform f_crtc_to_framebuffer;
+    struct pict_f_transform f_framebuffer_to_crtc;
+    PictFilterPtr   filter;
+    xFixed	    *params;
+    int		    nparams;
+    int		    filter_width;
+    int		    filter_height;
     Bool	    transform_in_use;
+    RRTransformRec  transform;
+    Bool	    transformPresent;
+    RRTransformRec  desiredTransform;
+    Bool	    desiredTransformPresent;
     /**
      * Bounding box in screen space
      */
     BoxRec	    bounds;
+    /**
+     * Panning:
+     * TotalArea: total panning area, larger than CRTC's size
+     * TrackingArea: Area of the pointer for which the CRTC is panned
+     * border: Borders of the displayed CRTC area which induces panning if the pointer reaches them
+     */
+    BoxRec          panningTotalArea;
+    BoxRec          panningTrackingArea;
+    INT16           panningBorder[4];
 };
 
 typedef struct _xf86OutputFuncs {
@@ -440,7 +466,7 @@ typedef struct _xf86OutputFuncs {
 } xf86OutputFuncsRec, *xf86OutputFuncsPtr;
 
 
-#define XF86_OUTPUT_VERSION 1
+#define XF86_OUTPUT_VERSION 2
 
 struct _xf86Output {
     /**
@@ -548,6 +574,10 @@ struct _xf86Output {
 #else
     void		*randr_output;
 #endif
+    /** Desired initial panning */
+    BoxRec          initialTotalArea;
+    BoxRec          initialTrackingArea;
+    INT16           initialBorder[4];
 };
 
 typedef struct _xf86CrtcConfigFuncs {
@@ -569,6 +599,8 @@ typedef struct _xf86CrtcConfigFuncs {
 	      int		height);
 } xf86CrtcConfigFuncsRec, *xf86CrtcConfigFuncsPtr;
 
+typedef void (*xf86_crtc_notify_proc_ptr) (ScreenPtr pScreen);
+
 typedef struct _xf86CrtcConfig {
     int			num_output;
     xf86OutputPtr	*output;
@@ -621,9 +653,12 @@ typedef struct _xf86CrtcConfig {
     /* wrap screen BlockHandler for rotation */
     ScreenBlockHandlerProcPtr	BlockHandler;
 
+    /* callback when crtc configuration changes */
+    xf86_crtc_notify_proc_ptr  xf86_crtc_notify;
+
 } xf86CrtcConfigRec, *xf86CrtcConfigPtr;
 
-extern int xf86CrtcConfigPrivateIndex;
+extern _X_EXPORT int xf86CrtcConfigPrivateIndex;
 
 #define XF86_CRTC_CONFIG_PTR(p)	((xf86CrtcConfigPtr) ((p)->privates[xf86CrtcConfigPrivateIndex].ptr))
 
@@ -631,11 +666,11 @@ extern int xf86CrtcConfigPrivateIndex;
  * Initialize xf86CrtcConfig structure
  */
 
-void
+extern _X_EXPORT void
 xf86CrtcConfigInit (ScrnInfoPtr				scrn,
 		    const xf86CrtcConfigFuncsRec	*funcs);
 
-void
+extern _X_EXPORT void
 xf86CrtcSetSizeRange (ScrnInfoPtr scrn,
 		      int minWidth, int minHeight,
 		      int maxWidth, int maxHeight);
@@ -643,121 +678,131 @@ xf86CrtcSetSizeRange (ScrnInfoPtr scrn,
 /*
  * Crtc functions
  */
-xf86CrtcPtr
+extern _X_EXPORT xf86CrtcPtr
 xf86CrtcCreate (ScrnInfoPtr		scrn,
 		const xf86CrtcFuncsRec	*funcs);
 
-void
+extern _X_EXPORT void
 xf86CrtcDestroy (xf86CrtcPtr		crtc);
 
 
 /**
  * Sets the given video mode on the given crtc
  */
-Bool
+
+extern _X_EXPORT Bool
+xf86CrtcSetModeTransform (xf86CrtcPtr crtc, DisplayModePtr mode, Rotation rotation,
+			  RRTransformPtr transform, int x, int y);
+
+extern _X_EXPORT Bool
 xf86CrtcSetMode (xf86CrtcPtr crtc, DisplayModePtr mode, Rotation rotation,
 		 int x, int y);
 
+extern _X_EXPORT void
+xf86CrtcSetOrigin (xf86CrtcPtr crtc, int x, int y);
+
 /*
  * Assign crtc rotation during mode set
  */
-Bool
-xf86CrtcRotate (xf86CrtcPtr crtc, DisplayModePtr mode, Rotation rotation);
+extern _X_EXPORT Bool
+xf86CrtcRotate (xf86CrtcPtr crtc);
 
 /*
  * free shadow memory allocated for all crtcs
  */
-void
+extern _X_EXPORT void
 xf86RotateFreeShadow(ScrnInfoPtr pScrn);
 
 /*
  * Clean up rotation during CloseScreen
  */
-void
+extern _X_EXPORT void
 xf86RotateCloseScreen (ScreenPtr pScreen);
 
 /**
  * Return whether any output is assigned to the crtc
  */
-Bool
+extern _X_EXPORT Bool
 xf86CrtcInUse (xf86CrtcPtr crtc);
 
 /*
  * Output functions
  */
-xf86OutputPtr
+extern _X_EXPORT xf86OutputPtr
 xf86OutputCreate (ScrnInfoPtr		    scrn,
 		  const xf86OutputFuncsRec  *funcs,
 		  const char		    *name);
 
-void
+extern _X_EXPORT void
 xf86OutputUseScreenMonitor (xf86OutputPtr output, Bool use_screen_monitor);
 
-Bool
+extern _X_EXPORT Bool
 xf86OutputRename (xf86OutputPtr output, const char *name);
 
-void
+extern _X_EXPORT void
 xf86OutputDestroy (xf86OutputPtr	output);
 
-void
+extern _X_EXPORT void
 xf86ProbeOutputModes (ScrnInfoPtr pScrn, int maxX, int maxY);
 
-void
+extern _X_EXPORT void
 xf86SetScrnInfoModes (ScrnInfoPtr pScrn);
 
 #ifdef RANDR_13_INTERFACE
-int
+# define ScreenInitRetType	int
 #else
-Bool
+# define ScreenInitRetType	Bool
 #endif
+
+extern _X_EXPORT ScreenInitRetType
 xf86CrtcScreenInit (ScreenPtr pScreen);
 
-Bool
+extern _X_EXPORT Bool
 xf86InitialConfiguration (ScrnInfoPtr pScrn, Bool canGrow);
 
-void
+extern _X_EXPORT void
 xf86DPMSSet(ScrnInfoPtr pScrn, int PowerManagementMode, int flags);
     
-Bool
+extern _X_EXPORT Bool
 xf86SaveScreen(ScreenPtr pScreen, int mode);
 
-void
+extern _X_EXPORT void
 xf86DisableUnusedFunctions(ScrnInfoPtr pScrn);
 
-DisplayModePtr
+extern _X_EXPORT DisplayModePtr
 xf86OutputFindClosestMode (xf86OutputPtr output, DisplayModePtr desired);
     
-Bool
+extern _X_EXPORT Bool
 xf86SetSingleMode (ScrnInfoPtr pScrn, DisplayModePtr desired, Rotation rotation);
 
 /**
  * Set the EDID information for the specified output
  */
-void
+extern _X_EXPORT void
 xf86OutputSetEDID (xf86OutputPtr output, xf86MonPtr edid_mon);
 
 /**
  * Return the list of modes supported by the EDID information
  * stored in 'output'
  */
-DisplayModePtr
+extern _X_EXPORT DisplayModePtr
 xf86OutputGetEDIDModes (xf86OutputPtr output);
 
-xf86MonPtr
+extern _X_EXPORT xf86MonPtr
 xf86OutputGetEDID (xf86OutputPtr output, I2CBusPtr pDDCBus);
 
 /**
  * Initialize dga for this screen
  */
 
-Bool
+extern _X_EXPORT Bool
 xf86DiDGAInit (ScreenPtr pScreen, unsigned long dga_address);
 
 /**
  * Re-initialize dga for this screen (as when the set of modes changes)
  */
 
-Bool
+extern _X_EXPORT Bool
 xf86DiDGAReInit (ScreenPtr pScreen);
 
 /*
@@ -765,13 +810,13 @@ xf86DiDGAReInit (ScreenPtr pScreen);
  * the information from the outputs
  */
 
-void
+extern _X_EXPORT void
 xf86CrtcSetScreenSubpixelOrder (ScreenPtr pScreen);
 
 /*
  * Get a standard string name for a connector type 
  */
-char *
+extern _X_EXPORT char *
 xf86ConnectorGetName(xf86ConnectorType connector);
 
 /*
@@ -779,7 +824,7 @@ xf86ConnectorGetName(xf86ConnectorType connector);
  * modes (used in EnterVT functions, or at server startup)
  */
 
-Bool
+extern _X_EXPORT Bool
 xf86SetDesiredModes (ScrnInfoPtr pScrn);
 
 /**
@@ -788,7 +833,7 @@ xf86SetDesiredModes (ScrnInfoPtr pScrn);
  *
  * Driver should call this from ScreenInit function
  */
-Bool
+extern _X_EXPORT Bool
 xf86_cursors_init (ScreenPtr screen, int max_width, int max_height, int flags);
 
 /**
@@ -798,25 +843,25 @@ xf86_cursors_init (ScreenPtr screen, int max_width, int max_height, int flags);
  * 
  * Driver should call this from crtc commit function.
  */
-void
+extern _X_EXPORT void
 xf86_reload_cursors (ScreenPtr screen);
 
 /**
  * Called from EnterVT to turn the cursors back on
  */
-void
+extern _X_EXPORT void
 xf86_show_cursors (ScrnInfoPtr scrn);
 
 /**
  * Called by the driver to turn cursors off
  */
-void
+extern _X_EXPORT void
 xf86_hide_cursors (ScrnInfoPtr scrn);
 
 /**
  * Clean up CRTC-based cursor code. Driver must call this at CloseScreen time.
  */
-void
+extern _X_EXPORT void
 xf86_cursors_fini (ScreenPtr screen);
 
 /*
@@ -825,7 +870,7 @@ xf86_cursors_fini (ScreenPtr screen);
  * wraps xf86XVClipVideoHelper()
  */
 
-Bool
+extern _X_EXPORT Bool
 xf86_crtc_clip_video_helper(ScrnInfoPtr pScrn,
 			    xf86CrtcPtr *crtc_ret,
 			    xf86CrtcPtr desired_crtc,
@@ -838,4 +883,28 @@ xf86_crtc_clip_video_helper(ScrnInfoPtr pScrn,
 			    INT32	width,
 			    INT32	height);
     
+extern _X_EXPORT xf86_crtc_notify_proc_ptr
+xf86_wrap_crtc_notify (ScreenPtr pScreen, xf86_crtc_notify_proc_ptr new);
+
+extern _X_EXPORT void
+xf86_unwrap_crtc_notify(ScreenPtr pScreen, xf86_crtc_notify_proc_ptr old);
+
+extern _X_EXPORT void
+xf86_crtc_notify(ScreenPtr pScreen);
+
+/**
+ * Panning
+ */
+Bool
+xf86_crtc_get_panning(ScrnInfoPtr pScrn,
+		      BoxPtr      totalArea,
+		      BoxPtr      TrackingArea,
+		      INT16      *border);
+
+Bool
+xf86_crtc_set_panning(ScrnInfoPtr pScrn,
+		      BoxPtr      totalArea,
+		      BoxPtr      TrackingArea,
+		      INT16      *border);
+
 #endif /* _XF86CRTC_H_ */
