@@ -41,9 +41,6 @@
 #include "xf86DDC.h"
 #include "randrstr.h"
 
-#define _XF86MISC_SERVER_
-#include <X11/extensions/xf86misc.h>
-
 #include "xf86Crtc.h"
 
 #ifdef USE_EXA
@@ -105,7 +102,9 @@ typedef enum
     TMDS_EXT     = 2,
     TMDS_LVTMA   = 3,
     TMDS_DDIA    = 4,
-    TMDS_UNIPHY  = 5
+    TMDS_UNIPHY  = 5,
+    TMDS_UNIPHY1 = 6,
+    TMDS_UNIPHY2 = 7
 } RADEONTmdsType;
 
 typedef enum
@@ -166,6 +165,8 @@ typedef struct
     Bool   valid;
     uint32_t mask_clk_reg;
     uint32_t mask_data_reg;
+    uint32_t a_clk_reg;
+    uint32_t a_data_reg;
     uint32_t put_clk_reg;
     uint32_t put_data_reg;
     uint32_t get_clk_reg;
@@ -176,15 +177,15 @@ typedef struct
     uint32_t put_data_mask;
     uint32_t get_clk_mask;
     uint32_t get_data_mask;
+    uint32_t a_clk_mask;
+    uint32_t a_data_mask;
+    int hw_line;
+    Bool hw_capable;
 } RADEONI2CBusRec, *RADEONI2CBusPtr;
 
 typedef struct _RADEONCrtcPrivateRec {
-#ifdef USE_XAA
-    FBLinearPtr rotate_mem_xaa;
-#endif
-#ifdef USE_EXA
-    ExaOffscreenArea *rotate_mem_exa;
-#endif
+    void *crtc_rotate_mem;
+    void *cursor_mem;
     int crtc_id;
     int binding;
     uint32_t cursor_offset;
@@ -206,6 +207,9 @@ typedef struct {
     int hpd_mask;
     RADEONI2CBusRec ddc_i2c;
     int igp_lane_info;
+    Bool shared_ddc;
+    int i2c_line_mux;
+    Bool load_detection;
 } RADEONBIOSConnector;
 
 typedef struct _RADEONOutputPrivateRec {
@@ -235,8 +239,10 @@ typedef struct _RADEONOutputPrivateRec {
     int               VSyncWidth;
     int               VBlank;
     int               Flags;            /* Saved copy of mode flags          */
-    int               PanelPwrDly;
     int               DotClock;
+    int               PanelPwrDly;
+    int               lvds_misc;
+    int               lvds_ss_id;
     RADEONTMDSPll     tmds_pll[4];
     RADEONRMXType     rmx_type;
     /* dvo */
@@ -263,6 +269,7 @@ typedef struct _RADEONOutputPrivateRec {
     int               output_id;
     int               devices;
     Bool enabled;
+    Bool shared_ddc;
 } RADEONOutputPrivateRec, *RADEONOutputPrivatePtr;
 
 struct avivo_pll_state {
@@ -310,8 +317,10 @@ struct avivo_grph_state {
     uint32_t x_end;
     uint32_t y_end;
 
+    uint32_t desktop_height;
     uint32_t viewport_start;
     uint32_t viewport_size;
+    uint32_t mode_data_format;
 };
 
 struct avivo_state
@@ -326,6 +335,7 @@ struct avivo_state
 
     uint32_t crtc_master_en;
     uint32_t crtc_tv_control;
+    uint32_t dc_lb_memory_split;
 
     struct avivo_pll_state pll1;
     struct avivo_pll_state pll2;
